/*
 * sysprof-live-pid.h
 *
 * Copyright 2024 Christian Hergert <chergert@redhat.com>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * SPDX-License-Identifier: GPL-3.0-or-later
 */

#pragma once

#include <glib.h>

G_BEGIN_DECLS

typedef struct _SysprofLiveProcess SysprofLiveProcess;

SysprofLiveProcess *sysprof_live_process_new       (GPid                pid);
SysprofLiveProcess *sysprof_live_process_ref       (SysprofLiveProcess *self);
void                sysprof_live_process_unref     (SysprofLiveProcess *self);
gboolean            sysprof_live_process_is_active (SysprofLiveProcess *self);
void                sysprof_live_process_add_map   (SysprofLiveProcess *self,
                                                    guint64             begin,
                                                    guint64             end,
                                                    guint64             offset,
                                                    guint64             inode,
                                                    const char         *filename);
guint               sysprof_live_process_unwind    (SysprofLiveProcess *self,
                                                    GPid                tid,
                                                    guint64             abi,
                                                    const guint8       *stack,
                                                    gsize               stack_len,
                                                    const guint64      *registers,
                                                    guint               n_registers,
                                                    guint64            *addresses,
                                                    guint               n_addresses);

G_END_DECLS
