
/* This file is generated by glib-mkenums, do not modify it. This code is licensed under the same license as the containing project. Note that it links to GLib, so must comply with the LGPL linking clauses. */

#pragma once

#include <glib-object.h>
#include "sysprof-version-macros.h"

G_BEGIN_DECLS

/* enumerations from "sysprof-callgraph.h" */

_SYSPROF_EXTERN
GType sysprof_callgraph_category_get_type (void);
#define SYSPROF_TYPE_CALLGRAPH_CATEGORY (sysprof_callgraph_category_get_type())

_SYSPROF_EXTERN
GType sysprof_callgraph_flags_get_type (void);
#define SYSPROF_TYPE_CALLGRAPH_FLAGS (sysprof_callgraph_flags_get_type())

/* enumerations from "sysprof-recording.h" */

_SYSPROF_EXTERN
GType sysprof_recording_phase_get_type (void);
#define SYSPROF_TYPE_RECORDING_PHASE (sysprof_recording_phase_get_type())

G_END_DECLS

/* Generated data ends here */

